#ifndef UsbDaqDevice_HH
#define UsbDaqDevice_HH

#include <unistd.h>

#include <cmath>
#include <cstdlib>
#include <iostream>

#include "TRandom.h"

#include "MpsUsbDaqRunData.hh"
#include "MpsUsbDaqConfigurationData.hh"
#include "MpsUsbDaqBunchTrainData.hh"
#include "MpsPcb1ConfigurationData.hh"
#include "MpsPcb1BunchTrainData.hh"
#include "MpsSensor1ConfigurationData.hh"
#include "MpsSensor1BunchTrainData.hh"


class UsbDaqDevice {

public:
  UsbDaqDevice() {
    _print=false;

    _usbDaqRunData.firmwareVersion(0xdeadbeef);

    gRandom->SetSeed(12345);
  }

  virtual ~UsbDaqDevice() {
  }

  // Run

  bool runReset() {
    return true;
  }

  bool readMpsUsbDaqRunData(MpsUsbDaqRunData &m) {
    m=_usbDaqRunData;
    return true;
  }

  // Configuration

  bool writeMpsUsbDaqConfigurationData(const MpsUsbDaqConfigurationData &m) {
    _usbDaqConfigurationData=m;

    _tMax=_usbDaqConfigurationData.numberOfBunchCrossings();

    std::cout << "UsbDaq paras = " << _tMax << std::endl;

    return true;
  }

  bool readMpsUsbDaqConfigurationData(MpsUsbDaqConfigurationData &m) {
    m=_usbDaqConfigurationData;
    return true;
  }

  bool writeMpsPcb1ConfigurationData(const MpsPcb1ConfigurationData &m) {
    _pcb1ConfigurationData=m;

    _threshold=_pcb1ConfigurationData.dac(0)/4096.0;
    _noiseRatePerPixel=pow(10.0,-12.0*_threshold);

    //if(_print)
    std::cout << "Pcb1 paras = " << _threshold << ", "
	      << _noiseRatePerPixel << std::endl;

    return true;
  }

  bool readMpsPcb1ConfigurationData(MpsPcb1ConfigurationData &m) {
    m=_pcb1ConfigurationData;
    return true;
  }

  bool loadMpsSensor1ConfigurationData(const MpsSensor1ConfigurationData &w,
				       MpsSensor1ConfigurationData &r) {
    r=_sensor1ConfigurationData;
    _sensor1ConfigurationData=w;
    return true;
  }

  // Bunch train

  bool bunchTrainReset() {
    if(_print) std::cout << "bunchTrainReset()" << std::endl;
    return true;
  }

  bool takeBunchTrain() {
    UtlTime start(true);

    // Find number of noise hits
    unsigned nNoise((unsigned)(gRandom->Poisson(168*168*_tMax*_noiseRatePerPixel)));
    if(nNoise>19*168*4) nNoise=19*168*4;
    if(_print) std::cout << "Number of noise hits = " << nNoise << std::endl;


    _vTrgs.clear();
    _vHits.clear();

    // Make list of noise hits
    for(unsigned i(0);i<nNoise;i++) {
      unsigned x((unsigned)(gRandom->Uniform()*168));
      unsigned y((unsigned)(gRandom->Uniform()*168));
      unsigned t((unsigned)(gRandom->Uniform()*_tMax));
      
      if(!_sensor1ConfigurationData.mask(x,y)) {
	MpsSensor1Hit hit;
	hit.pixelX(x);
	hit.pixelY(y);
	hit.timeStamp(t);
	_vHits.push_back(hit);
      }
    }

    if(_print) std::cout << "Number of stored noise hits = " << _vHits.size() << std::endl;
    

    /*
    unsigned n(0);
    UtlPack *d((UtlPack*)(&m+1));
    for(unsigned j(0);j<168;j++) {
      for(unsigned k(0);k<_tMax;k++) {
	for(unsigned i(0);i<168;i+=6) {
	  if(__vHits[i][j][k]) {
	    unsigned col(i/42);
	    unsigned grp((i%42)/7);

	    d[n].bits(0,2,grp);
	    for(unsigned l(i);l<i+6;l++) d[n].bit(3+l,__vHits[l][j][k]);
	    d[n].bits(9,17,j);
	    d[n].bits(18,30,k);
	    n++;
	  }
	}
      }
    }
    */

    // Add physics hits

    double efficiency(2.0-2.0*_threshold);
    if(efficiency>gRandom->Uniform()) {

      // Source
      /*
      unsigned x((unsigned)(gRandom->Gaus(42.0+21.0,1.0)));
      unsigned y((unsigned)(gRandom->Gaus(84.0+21.0,1.0)));
      unsigned t((unsigned)(gRandom->Uniform()*_tMax));
      */

      // Laser
      double xl(0.001*_laserX);
      double yl(0.001*_laserY);

      double xs(xl);
      double ys(yl);
      
      double xn(83.5+(xs/0.05));
      double yn(83.5+(ys/0.05));

      unsigned x((unsigned)(gRandom->Gaus(xn,1.0)));
      unsigned y((unsigned)(gRandom->Gaus(yn,1.0)));
      unsigned t((unsigned)(gRandom->Uniform()*_tMax));
      
      if(x<168 && y<168 && !_sensor1ConfigurationData.mask(x,y)) {
	MpsSensor1Hit hit;
	hit.pixelX(x);
	hit.pixelY(y);
	hit.timeStamp(t);
	
	if(t  >    0) _vTrgs.push_back(((t-1)<<16)+0x1);
	_vTrgs.push_back(((t  )<<16)+0x3);
	if(t+1<_tMax) _vTrgs.push_back(((t+1)<<16)+0x2);
	if(t+2<_tMax) _vTrgs.push_back(((t+2)<<16)+0x2);
	
	// Check no noise hit already
	bool alreadyHit(false);
	for(unsigned i(0);i<_vHits.size();i++) {
	  if(_vHits[i].pixelX()==hit.pixelX() && 
	     _vHits[i].pixelY()==hit.pixelY() &&
	     _vHits[i].timeStamp()==hit.timeStamp()) alreadyHit=true;
	}
	if(!alreadyHit) _vHits.push_back(hit);
      }
    }
    
    if(_print) std::cout << "Number of total hits = "
			 << _vHits.size() << std::endl;



    UtlTime finish(true);
    UtlTimeDifference diff(finish-start);
    if(diff.seconds()==0) {
      if(diff.microseconds()>=0 && diff.microseconds()<90000) {
	usleep(90000-diff.microseconds());
      }
    }


    return true;
  }

  bool readMpsUsbDaqBunchTrainData(MpsUsbDaqBunchTrainData &m) {
    m.numberOfTries(1);
    m.trgVector(_vTrgs);
    return true;
  }

  bool readMpsPcb1BunchTrainData(MpsPcb1BunchTrainData &m) {
    m.temperature((unsigned)gRandom->Gaus(20.0,2.0));
    return true;
  }

  bool readMpsSensor1BunchTrainData(MpsSensor1BunchTrainData &m) {
    m.finalBits(0xaaaa);
    m.hitVector(_vHits);
    return true;
  }

  static void coordinates(int x, int y) {
    _laserX=x;
    _laserY=y;
  }

private:
  bool _print;
 
  MpsUsbDaqRunData _usbDaqRunData;

  MpsUsbDaqConfigurationData  _usbDaqConfigurationData;
  MpsPcb1ConfigurationData    _pcb1ConfigurationData;
  MpsSensor1ConfigurationData _sensor1ConfigurationData;

  unsigned _tMax;
  double _threshold;
  double _noiseRatePerPixel;

  // Bunch train output
  std::vector<unsigned> _vTrgs;
  std::vector<MpsSensor1Hit> _vHits;


  // CLUDGE!!!
  static int _laserX;
  static int _laserY;
};

int UsbDaqDevice::_laserX(0);
int UsbDaqDevice::_laserY(0);

#endif
